from typing import Optional, List, Tuple

from cosmpy.crypto.address import Address

from xchainpy2_client import XcTx, TxType, TokenTransfer
from xchainpy2_utils import Asset, AssetATOM, Chain, Amount, key_attr_getter, parse_iso_date
from .const import COSMOS_DENOM
from .models import TxLoadException, load_logs, TxLog


def is_msg_multi_send(msg):
    return key_attr_getter(msg, 'inputs') and key_attr_getter(msg, 'outputs')


def is_msg_send(msg):
    return key_attr_getter(msg, 'from_address') and \
        key_attr_getter(msg, 'to_address') and \
        key_attr_getter(msg, 'amount')


def get_denom(asset: Asset) -> str:
    """
    Get denomination from Asset - currently `ATOM` supported only
    :param asset:
    :return: The denomination of the given asset.
    """
    if asset == AssetATOM:
        return COSMOS_DENOM
    return ''


def get_asset(denom: str, native_denom=COSMOS_DENOM, native_asset=AssetATOM) -> Optional[Asset]:
    """
    Get Asset from denomination - currently `ATOM` supported only
    :param denom:
    :param native_asset:
    :param native_denom:
    :return: The asset of the given denomination.
    """
    if denom == native_denom:
        return native_asset

    # IBC assets
    if denom.startswith('ibc/'):
        # Note: Don't use `assetFromString` here, it will interpret `/` as synth
        return Asset(
            Chain.Cosmos.value,
            symbol=denom.split('/', maxsplit=2)[1],
            # TODO (xchain-contributors)
            # Get readable ticker for IBC assets from denom #600 https://github.com/xchainjs/xchainjs-lib/issues/600
            # At the meantime ticker will be empty
            contract='',
            synth=False
        )

    return None


def get_coin_amount(coins, decimals) -> Amount:
    return sum(
        (Amount.from_base(int(key_attr_getter(coin, 'amount')), decimals) for coin in coins),
        Amount.from_base(0, decimals)
    )


def get_coins_by_asset(coins, search_asset: Asset, native_denom: str, native_asset: Asset):
    return [
        coin for coin in coins
        if get_asset(key_attr_getter(coin, 'denom'), native_denom, native_asset) == search_asset
    ]


def parse_transfer_log(log: TxLog, decimals, filter_address, native_denom: str, native_asset: Asset) \
        -> List[TokenTransfer]:
    transfer_data_list = []
    sender, recipient, amount, asset = None, None, None, None
    for event in log.events:
        if event.type == 'transfer':
            for i, attribute in enumerate(event.attributes):
                if attribute.key == 'sender':
                    sender = attribute.value
                elif attribute.key == 'recipient':
                    recipient = attribute.value
                elif attribute.key == 'amount':
                    amount_int, asset = parse_cosmos_amount(attribute.value)
                    amount = Amount.from_base(amount_int, decimals)
                    if asset == native_denom:
                        asset = native_asset
                    else:
                        asset = Asset.from_string(asset.upper())

                # ready to append
                if sender and recipient and amount and asset:
                    transfer_data_list.append(
                        TokenTransfer(
                            sender, recipient, amount, asset,
                            tx_hash='', outbound=(sender == filter_address)
                        ))

                    sender, recipient, amount, asset = None, None, None, None  # reset

    if filter_address:
        transfer_data_list = [
            data for data in transfer_data_list
            if data.from_address == filter_address or data.to_address == filter_address
        ]

    return transfer_data_list


def parse_tx_response_json(j: dict, tx_id: str, address: str, decimals: int,
                           native_denom: str, native_asset: Asset) -> XcTx:
    response = j.get('tx_response')
    if not response:
        raise TxLoadException(f'Failed to get transaction logs (tx-hash: ${tx_id}): no tx_response')

    code = response.get('code')
    is_successful = code == 0
    if not is_successful:
        raise TxLoadException(f'Code is not 0 ({code}) for tx-hash: ${tx_id}')

    messages = j['tx']['body']['messages']

    if len(messages) > 1:
        raise TxLoadException(f'Multiple messages are not supported yet (tx-hash: ${tx_id})')
    elif not messages:
        raise TxLoadException(f'No messages found (tx-hash: ${tx_id})')

    address = address or messages[0].get('signer') or messages[0].get('from_address')

    logs = load_logs(response.get('logs'))
    if not logs:
        raise TxLoadException(f'Failed to get transaction logs (tx-hash: ${tx_id})')

    if len(logs) > 1:
        raise TxLoadException(f'Multiple logs are not supported yet (tx-hash: ${tx_id})')

    log = logs[0]

    transfers_events = log.find_events('transfer')
    message_event = log.find_event('message')
    if not transfers_events and not message_event:
        raise TxLoadException(f'Invalid transaction data, no transfer/no message (tx-hash: ${tx_id})')

    memo = ''
    if (tx_j := j.get('tx')) and (body := tx_j.get('body')):
        memo = body['memo']

    tx_date = parse_iso_date(response['timestamp'])
    tx_hash = response['txhash']
    height = int(response['height'])

    transfers = parse_transfer_log(log, decimals, address, native_denom, native_asset)

    return XcTx(
        asset=transfers[0].asset if transfers else None,
        transfers=transfers,
        date=tx_date,
        type=TxType.TRANSFER,
        hash=tx_hash,
        height=height,
        memo=memo,
        original=j,
    )


def parse_cosmos_amount(amount: str) -> Tuple[int, str]:
    """
    Parse COSMOS amount
    into integer amount and string denomination
    :param amount: a string with format "123asset", like "100rune"
    :return: (int, str)
    """
    for i, char in enumerate(amount):
        if not char.isdigit():
            pos = i
            break
    else:
        raise ValueError(f'Invalid format: {amount!r} Must be like "10rune"')

    return (
        int(amount[:pos]),
        amount[pos:]
    )


def parse_cosmos_amounts(value: str) -> List[Tuple[int, str]]:
    """
    Parse COSMOS amounts from string like "1234rune,1234bnb"
    :param value: Cosmos SDK amount string
    :return: List of (value, asset) pairs
    """
    if not value:
        return []
    items = value.split(',')
    return [parse_cosmos_amount(item.strip()) for item in items]


def convert_address_for_msg(address, prefix: str) -> bytes:
    return bytes(Address(address, prefix))



from xchainpy2_cosmos import RPCResponse
from xchainpy2_utils.testing_utils import load_example_json


def test_rpc_response():
    r = RPCResponse.from_rpc_json({
        "jsonrpc": "foo",
        "id": 100,
        "result": "good"
    })
    assert r.result == "good"
    assert r.id == 100
    assert r.jsonrpc == "foo"


def test_parse_tx_response():
    tx_hash = '40BD2FCF41252AD7EAEE313989DC9CB10A120CD384804DD0E9D87F1B388243BB'
    raw_json = load_example_json(__file__, f'samples/{tx_hash}.json')

    # assert xctx.asset == AssetRUNE
    # assert xctx.hash == tx_hash
    # assert xctx.height == 11646005
    # assert xctx.type == TxType.TRANSFER
    # assert xctx.date == datetime.datetime(2023, 7, 10, 15, 49, 56)
    #
    # todo


# The decimal for cosmos chain.
from typing import Optional, Callable

from xchainpy2_client import Fees, FeeType, FeeOption, ExplorerProvider
from xchainpy2_utils import Amount, NetworkType

COSMOS_DECIMAL = 6

# Default gas limit
# As same as definition in Cosmosstation's web wallet
DEFAULT_GAS_LIMIT = '200000'

# Default fee
# As same as definition in Cosmosstation's web wallet
DEFAULT_FEE = Amount.from_base(5000, COSMOS_DECIMAL)

# Chain identifier for Cosmos chain
GAIA_CHAIN_KEY = 'GAIA'

# Cosmos denomination
COSMOS_DENOM = 'uatom'

# Default derivation path for Cosmos chain
DEFAULT_DERIVATION_PATH = "44'/118'/0'/0/"

COSMOS_ROOT_DERIVATION_PATHS = {
    NetworkType.MAINNET: DEFAULT_DERIVATION_PATH,
    NetworkType.TESTNET: DEFAULT_DERIVATION_PATH,
    NetworkType.STAGENET: DEFAULT_DERIVATION_PATH,
}

# Prefix
COSMOS_ADDR_PREFIX = 'cosmos'


def get_default_fees() -> Fees:
    return Fees(
        type=FeeType.FLAT_FEE,
        fees={
            FeeOption.FAST: Amount.from_base(750, COSMOS_DECIMAL),
            FeeOption.FASTEST: Amount.from_base(2500, COSMOS_DECIMAL),
            FeeOption.AVERAGE: Amount.from_base(0, COSMOS_DECIMAL),
        }
    )


CLIENT_URL_KEPLR = 'https://lcd-cosmoshub.keplr.app'
CLIENT_URL_COSMOSTATION = 'https://lcd-cosmos.cosmostation.io/'
# The only one that works for me
CLIENT_URL_COSMOS_DIRECTORY = 'https://rest.cosmos.directory/cosmoshub'

DEFAULT_CLIENT_URLS = {
    NetworkType.MAINNET: CLIENT_URL_KEPLR,
    NetworkType.STAGENET: CLIENT_URL_KEPLR,
    # Note: In case anyone facing into CORS issue, try the following URLs
    #   // https://lcd-cosmos.cosmostation.io/
    #   // https://lcd-cosmoshub.keplr.app/
    #   // @see (Discord #xchainjs) https://discord.com/channels/838986635756044328/988096545926828082/988103739967688724
    NetworkType.TESTNET: 'https://rest.sentry-02.theta-testnet.polypore.xyz',
}

COSMOS_CHAIN_IDS = {
    NetworkType.MAINNET: 'cosmoshub-4',
    NetworkType.STAGENET: 'cosmoshub-4',
    NetworkType.TESTNET: 'theta-testnet-001',
}

TEST_EXPLORER_URL = 'https://explorer.theta-testnet.polypore.xyz/{path}'
BIG_DIPPER_EXPLORER_URL = 'https://bigdipper.live/cosmos/{path}'
MINT_SCAN_EXPLORER_URL = 'https://www.mintscan.io/cosmos/{path}'


def make_explorer(path, acc_subpath, tx_subpath):
    return ExplorerProvider(
        path.format(path=''),
        path.format(path=acc_subpath),
        path.format(path=tx_subpath),
    )


def make_explorer_std(path):
    return make_explorer(path, 'accounts/{address}', 'transactions/{tx_id}')


def make_explorer_mint_scan():
    return make_explorer(MINT_SCAN_EXPLORER_URL, 'account/{address}', 'txs/{tx_id}')


MINT_SCAN_EXPLORER_PROVIDER = {
    NetworkType.MAINNET: make_explorer_mint_scan(),
    NetworkType.STAGENET: make_explorer_mint_scan(),
    NetworkType.TESTNET: make_explorer_std(TEST_EXPLORER_URL),
}

BIG_DIPPER_EXPLORER_PROVIDER = {
    NetworkType.MAINNET: make_explorer_std(BIG_DIPPER_EXPLORER_URL),
    NetworkType.STAGENET: make_explorer_std(BIG_DIPPER_EXPLORER_URL),
    NetworkType.TESTNET: make_explorer_std(TEST_EXPLORER_URL),
}

DEFAULT_EXPLORER_PROVIDER = MINT_SCAN_EXPLORER_PROVIDER

TxFilterFunc = Optional[Callable[[object], bool]]

MAX_TX_COUNT_PER_PAGE = 100

MAX_TX_COUNT_PER_FUNCTION_CALL = 500

MAX_PAGES_PER_FUNCTION_CALL = 15

DEFAULT_REST_USER_AGENT = \
    'Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36'



import asyncio
import logging
from datetime import datetime
from operator import itemgetter
from typing import Optional, List, Union
from urllib.parse import urlencode

from aiohttp import ClientSession
from cosmpy.aerial.client import LedgerClient, Account, create_bank_send_msg, prepare_and_broadcast_basic_transaction
from cosmpy.aerial.config import NetworkConfig
from cosmpy.aerial.tx import Transaction
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PrivateKey, PublicKey
from cosmpy.protos.cosmos.tx.v1beta1.service_pb2 import BroadcastTxRequest, BroadcastMode
from math import ceil

from xchainpy2_client import XChainClient, RootDerivationPaths, FeeBounds, XcTx, \
    Fees, TxPage, AssetInfo, FeeType, FeeOption
from xchainpy2_client.fees import single_fee
from xchainpy2_crypto import derive_private_key, create_address
from xchainpy2_utils import Chain, NetworkType, CryptoAmount, RUNE_DECIMAL, Asset, Amount, AssetATOM, \
    unique_by_key, batched, NINE_REALMS_CLIENT_HEADER, XCHAINPY_IDENTIFIER, flatten
from .const import DEFAULT_CLIENT_URLS, DEFAULT_EXPLORER_PROVIDER, COSMOS_ROOT_DERIVATION_PATHS, COSMOS_ADDR_PREFIX, \
    COSMOS_CHAIN_IDS, COSMOS_DECIMAL, TxFilterFunc, MAX_PAGES_PER_FUNCTION_CALL, MAX_TX_COUNT_PER_PAGE, \
    MAX_TX_COUNT_PER_FUNCTION_CALL, COSMOS_DENOM, DEFAULT_FEE, DEFAULT_GAS_LIMIT, DEFAULT_REST_USER_AGENT
from .models import TxHistoryResponse, TxLoadException
from .utils import parse_tx_response_json

logger = logging.getLogger(__name__)


class CosmosGaiaClient(XChainClient):
    def __init__(self,
                 network=NetworkType.MAINNET,
                 phrase: Optional[str] = None,
                 private_key: Union[str, bytes, callable, None] = None,
                 fee_bound: Optional[FeeBounds] = None,
                 root_derivation_paths: Optional[RootDerivationPaths] = None,
                 client_urls=DEFAULT_CLIENT_URLS,
                 chain_ids=COSMOS_CHAIN_IDS,
                 explorer_providers=DEFAULT_EXPLORER_PROVIDER,
                 wallet_index=0,
                 ):
        """
        Initialize CosmosClient
        :param network: Network type. Default is `NetworkType.MAINNET`
        :param phrase: Mnemonic phrase
        :param private_key: Private key (if you want to use a private key instead of a mnemonic phrase)
        :param fee_bound: Fee bound structure. See: FeeBounds
        :param root_derivation_paths: Dictionary of derivation paths for each network type. See: ROOT_DERIVATION_PATHS
        :param client_urls: Dictionary of client urls for each network type.
        :param chain_ids: Dictionary of chain ids for each network type.
        :param explorer_providers: Dictionary of explorer providers for each network type.
        :param wallet_index: int (default 0)
        """
        root_derivation_paths = root_derivation_paths.copy() \
            if root_derivation_paths else COSMOS_ROOT_DERIVATION_PATHS.copy()
        super().__init__(Chain.Cosmos, network, phrase, private_key, fee_bound, root_derivation_paths, wallet_index)

        self.explorers = explorer_providers

        if isinstance(client_urls, str):
            client_urls = {NetworkType.MAINNET: client_urls}

        self.client_urls = client_urls.copy() if client_urls else DEFAULT_CLIENT_URLS.copy()
        self.chain_ids = chain_ids.copy() if chain_ids else COSMOS_CHAIN_IDS.copy()

        self._gas_asset = AssetATOM
        self._prefix = COSMOS_ADDR_PREFIX

        self._denom = COSMOS_DENOM
        self._decimal = COSMOS_DECIMAL
        self._gas_limit = DEFAULT_GAS_LIMIT

        self._client: Optional[LedgerClient] = None
        self._recreate_client()

        self._wallet: Optional[LocalWallet] = None
        self._make_wallet()

        self.cache = None
        self.tx_responses = {}

    def get_client(self) -> LedgerClient:
        """
        Please use this getter to obtain LedgerClient for specific Cosmos calls like delegation, staking etc.
        Underlying library is cosmpy (https://github.com/fetchai/cosmpy)
        :return: LedgerClient
        """
        if not self._client:
            self._recreate_client()
        return self._client

    @property
    def server_url(self):
        return self.client_urls[self.network]

    @property
    def rpc_url(self):
        return self.server_url

    @property
    def rest_session(self) -> ClientSession:
        # noinspection PyProtectedMember
        return self._client.auth._rest_api._session

    def patch_client(self, user_agent=DEFAULT_REST_USER_AGENT, identifier9r=XCHAINPY_IDENTIFIER):
        headers = self.rest_session.headers
        headers['User-Agent'] = user_agent
        headers[NINE_REALMS_CLIENT_HEADER] = identifier9r

    def _recreate_client(self):
        # Guard for preventing early client creation before all necessary fields are set.
        if not hasattr(self, '_denom'):
            return

        rest_url = f'rest+{self.server_url}'
        self._client = LedgerClient(NetworkConfig(
            chain_id=self.chain_ids[self.network],
            url=rest_url,
            fee_denomination=self._denom,
            staking_denomination=self._denom,
            fee_minimum_gas_price=0,
        ))

    def set_chain_id(self, chain_id: str):
        """
        Set/update the current chain id.
        :param chain_id:
        :return: None
        """
        # dirty check to avoid using and re-creation of same data
        if chain_id == self.chain_ids[self.network]:
            return
        self.chain_ids[self.network] = chain_id
        self._recreate_client()

    def get_chain_id(self):
        return self.chain_ids[self.network]

    def set_network(self, network: NetworkType):
        """
        Set/update the current network.
        :param network:
        :return: None
        """
        # dirty check to avoid using and re-creation of same data
        if network == self.network:
            return
        super().set_network(network)
        self._recreate_client()
        self._make_wallet()

    def get_network(self):
        return self.network

    def purge_client(self):
        super().purge_client()

    def set_phrase(self, phrase: str, wallet_index: int = 0):
        super().set_phrase(phrase, wallet_index)
        self._make_wallet()

    def validate_address(self, address: str) -> bool:
        """
        Validates a Cosmos address.
        :param address: Address string
        :return: True if valid, False otherwise.
        """

        if not address:
            return False

        address = str(address).lower()

        if not address.startswith(self._prefix):
            return False

        try:
            Address(address)
            return True
        except Exception:
            return False

    def get_address(self) -> str:
        """
        Get the address for the given wallet index.
        :return: string address
        """
        pub_key = self.get_public_key().public_key_bytes
        return create_address(pub_key, self._prefix)

    def get_private_key(self) -> str:
        """
        Get the private key for the given wallet index.
        :return:
        """
        if self.pk_hex:
            return self.pk_hex
        else:
            return derive_private_key(
                self.phrase,
                self.get_full_derivation_path(self.wallet_index)
            ).hex()

    def get_public_key(self) -> PublicKey:
        return self.get_private_key_cosmos().public_key

    def get_private_key_cosmos(self) -> PrivateKey:
        pk = self.get_private_key()
        return PrivateKey(bytes.fromhex(pk))

    async def get_balance(self, address: str = '') -> List[CryptoAmount]:
        """
        Get the balance of a given address.
        :param address: By default, it will return the balance of the current wallet. (optional)
        :return:
        """
        if not address:
            address = self.get_address()

        address = Address(address, prefix=self.prefix)

        balances = await asyncio.get_event_loop().run_in_executor(
            None,
            self._client.query_bank_all_balances,
            address
        )

        our_balances = [
            CryptoAmount(
                Amount.from_base(balance.amount, self._decimal),
                self.parse_denom_to_asset(balance.denom)
            )
            for balance in balances
        ]

        return our_balances

    def parse_denom_to_asset(self, denom: str) -> Asset:
        if denom == self._denom:
            return self._gas_asset
        else:
            return Asset(self.chain.value, denom.upper())

    async def get_account(self, address: str = None) -> Optional[Account]:
        """
        Get the account information for the given address: number and sequence.
        It there is no account, it will return None.
        It will throw an exception for special addresses (like Reserve)
        :param address: By default, it will return the account of the current wallet. (optional)
        :return:
        """
        if address is None:
            address = self.get_address()

        address = Address(address)

        try:
            account = await asyncio.get_event_loop().run_in_executor(
                None,
                self._client.query_account,
                address
            )
            return account
        except RuntimeError as e:
            if 'NotFound' in str(e):
                return
            raise e

    async def get_transactions(self, address: str = None,
                               offset: int = 0,
                               limit: int = 10,
                               start_time: Optional[datetime] = None,
                               end_time: Optional[datetime] = None,
                               asset: Optional[Asset] = None,
                               filter_function: TxFilterFunc = None,
                               batch_size=10,
                               batch_delay_sec=2) -> TxPage:
        if not address:
            address = self.get_address()

        message_action = None

        address = address if address else self.get_address()
        tx_min_height = None
        tx_max_height = None

        if limit + offset > MAX_PAGES_PER_FUNCTION_CALL * MAX_TX_COUNT_PER_PAGE:
            raise ValueError(
                f"limit plus offset can not be grater than {MAX_PAGES_PER_FUNCTION_CALL * MAX_TX_COUNT_PER_PAGE}")

        if limit > MAX_TX_COUNT_PER_FUNCTION_CALL:
            raise ValueError(f"Maximum number of transaction per call is {MAX_TX_COUNT_PER_FUNCTION_CALL}")

        pages_number = int(ceil((limit + offset) / MAX_TX_COUNT_PER_PAGE))

        all_tx_incoming_history = []
        all_tx_outgoing_history = []

        for page in range(1, pages_number + 1):
            call = self.search_tx_from_rpc(
                message_action=message_action,
                transfer_recipient=address,
                page=page,
                limit=limit,
                tx_max_height=tx_max_height,
                tx_min_height=tx_min_height,
                rpc_endpoint=self.rpc_url,
            )
            all_tx_incoming_history.append(call)

            call = self.search_tx_from_rpc(
                message_action=message_action,
                transfer_sender=address,
                page=page,
                limit=limit,
                tx_max_height=tx_max_height,
                tx_min_height=tx_min_height,
                rpc_endpoint=self.rpc_url,
            )
            all_tx_outgoing_history.append(call)

        # todo: add semaphore to limit the number of concurrent requests
        incoming_results = await asyncio.gather(*all_tx_incoming_history)
        outgoing_results = await asyncio.gather(*all_tx_outgoing_history)

        total_count = 0
        if incoming_results:
            total_count += int(incoming_results[0]['result']['total_count'])

        if outgoing_results:
            total_count += int(outgoing_results[0]['result']['total_count'])

        results = [results['result']['txs'] for results in incoming_results + outgoing_results]
        results = flatten(results)

        results = unique_by_key(results, itemgetter('hash'))

        results.sort(key=lambda x: int(x['height']), reverse=True)

        if filter_function:
            results = filter(filter_function, results)

        all_txs = []
        for batch in batched(results, batch_size):
            txs = await asyncio.gather(*[self.get_transaction_data(tx['hash']) for tx in batch])
            all_txs.extend(txs)
            await asyncio.sleep(batch_delay_sec)

        return TxPage(
            txs=all_txs,
            total=total_count,
        )

    async def search_tx(self, message_action=None, message_sender=None, offset=0, limit=50):
        if not message_action and not message_sender:
            raise ValueError('One of message_action or message_sender must be specified')

        events_param = []
        if message_action is not None:
            events_param.append(f"message.action='{message_action}'")
        if message_sender is not None:
            events_param.append(f"message.sender='{message_sender}'")

        query_parameter = {'events': ','.join(events_param)}

        if offset is not None:
            query_parameter['pagination.offset'] = offset
        if limit is not None:
            query_parameter['pagination.limit'] = limit

        url = f"{self.server_url}/cosmos/tx/v1beta1/txs?{urlencode(query_parameter)}"
        j = await self._get_json(url)
        return TxHistoryResponse.from_rpc_json(j)

    async def search_tx_from_rpc(
            self,
            message_action=None,
            message_sender=None,
            transfer_sender=None,
            transfer_recipient=None,
            page=None,
            limit=None,
            tx_min_height=None,
            tx_max_height=None,
            rpc_endpoint=None
    ):
        query_parameter = []
        if message_action is not None:
            query_parameter.append(f"message.action='{message_action}'")
        if message_sender is not None:
            query_parameter.append(f"message.sender='{message_sender}'")
        if transfer_sender is not None:
            query_parameter.append(f"transfer.sender='{transfer_sender}'")
        if transfer_recipient is not None:
            query_parameter.append(f"transfer.recipient='{transfer_recipient}'")
        if tx_min_height is not None:
            query_parameter.append(f"tx.height>='{tx_min_height}'")
        if tx_max_height is not None:
            query_parameter.append(f"tx.height<='{tx_max_height}'")

        search_parameters = {}
        if query_parameter:
            query = ' AND '.join(query_parameter)
            search_parameters['query'] = f'"{query}"'

        if page is not None:
            search_parameters['page'] = page
        if limit is not None:
            search_parameters['per_page'] = limit
        search_parameters['order_by'] = '"desc"'

        params_joined = '&'.join(f'{k}={v}' for k, v in search_parameters.items())

        url = f"{rpc_endpoint}/tx_search?{params_joined}"

        return await self._get_json(url)

    def url_to_fetch_tx_data(self, tx_id, server_url=None):
        server_url = server_url or self.server_url
        return f"{server_url}/cosmos/tx/v1beta1/txs/{tx_id}"

    async def get_transaction_data_cosmos(self, tx_id: str) -> dict:
        """
        Get the transaction data for the given transaction id.
        :param tx_id:
        :return: raw dict object
        """
        url = self.url_to_fetch_tx_data(tx_id)
        return await self._get_json(url)

    async def get_transaction_data(self, tx_id: str, our_address: str = '') -> Optional[XcTx]:
        """
        Get the transaction data for the given transaction id.
        :param tx_id:
        :return:
        """
        j = await self.get_transaction_data_cosmos(tx_id)

        return parse_tx_response_json(
            j, tx_id, our_address, self._decimal, self._denom, self._gas_asset
        )

    async def get_transaction_data_raw(self, tx_id: str) -> dict:
        url = self.url_to_fetch_tx_data(tx_id)
        return await self._get_json(url)

    async def get_fees(self, cache=None, tc_fee_rate=None) -> Fees:
        """
        Returns fees.
        It tries to get chain fees from THORChain `inbound_addresses` first
        :param cache: THORChainCache from the query module
        :param tc_fee_rate: You can externally pass the fee rate having 8 decimals. (optional)
        :return:
        """
        if tc_fee_rate is not None:
            return single_fee(FeeType.FLAT_FEE, DEFAULT_FEE)

        tc_fee_rate = await cache.get_fee_rates(self.chain)

        # convert decimal: 1e8 (THORChain) to 1e6 (COSMOS)
        # Similar to `fromCosmosToThorchain` in THORNode
        decimal_diff = self._decimal - RUNE_DECIMAL
        fee_rate = Amount.from_base(tc_fee_rate * 10 ** decimal_diff, self._decimal)
        return single_fee(FeeType.FLAT_FEE, fee_rate)

    async def transfer(self, what: CryptoAmount,
                       recipient: str,
                       memo: Optional[str] = None,
                       fee_rate: Optional[int] = None,
                       check_balance: bool = True) -> str:
        """
        Transfer coins.
        :param check_balance: Check balance before transfer. Default is True.
        :param what: CryptoAmount (amount and asset to transfer)
        :param recipient: str recipient address
        :param memo: str
        :param fee_rate: int
        :return: str tx hash
        """
        self._throw_if_empty_phrase()

        if check_balance:
            await self.check_balance(str(self._wallet.address()), what)

        tx = self.build_transfer_tx(what, recipient)

        response = await asyncio.get_event_loop().run_in_executor(
            None,
            prepare_and_broadcast_basic_transaction,
            self._client,
            tx,
            self._wallet,
            None,  # account
            int(self._gas_limit),
            memo
        )

        self._save_last_response(response.tx_hash, response)

        return response.tx_hash

    def build_transfer_tx(self, what: CryptoAmount, recipient: str) -> Transaction:
        tx = Transaction()
        tx.add_message(
            create_bank_send_msg(self._wallet.address(), Address(recipient),
                                 what.amount.internal_amount, self.get_denom(what.asset))
        )
        return tx

    async def broadcast_tx(self, tx_hex: str) -> str:
        broadcast_req = BroadcastTxRequest(
            tx_bytes=tx_hex.encode('utf-8'), mode=BroadcastMode.BROADCAST_MODE_SYNC
        )

        # broadcast the transaction
        resp = await asyncio.get_event_loop().run_in_executor(
            None,
            self._client.txs.BroadcastTx,
            broadcast_req)
        tx_digest = resp.tx_response.txhash

        # check that the response is successful
        initial_tx_response = self._client._parse_tx_response(resp.tx_response)
        initial_tx_response.ensure_successful()

        self._save_last_response(tx_digest, initial_tx_response)

        return tx_digest

    async def fetch_chain_id(self, server='') -> str:
        """
        Helper to get Cosmos' chain id
        :return:
        """
        url = f"{server or self.server_url}/node_info"
        j = await self._get_json(url)
        return j['node_info']['network']

    async def _get_json(self, url):
        logger.debug(f"GET {url}")
        response = await asyncio.get_event_loop().run_in_executor(
            None,
            self._client.txs.rest_client._session.get,
            url,
        )

        if response.status_code != 200:
            raise TxLoadException(f"Error getting {url!r}: code={response.status_code}; {response.text!r}")
        return response.json()

    @property
    def sdk_client(self) -> LedgerClient:
        return self._client

    @property
    def prefix(self) -> str:
        return self._prefix

    async def check_balance(self, address, amount: CryptoAmount):
        fees = await self.get_fees(self.cache)
        fee = fees.fees[FeeOption.AVERAGE]

        balances = await self.get_balance(address)

        asset_balance = None
        native_balance = None

        for balance in balances:
            if balance.asset == amount.asset:
                asset_balance = balance
            if balance.asset == self._gas_asset:
                native_balance = balance

        is_native = amount.asset == self._gas_asset
        extra_fee = fee if is_native else Amount.from_base(0, self._decimal)

        if asset_balance is None or asset_balance.amount.as_base < amount.amount.as_base + extra_fee.as_base:
            raise ValueError(f"Insufficient funds: {amount.amount} {amount.asset}")

        if native_balance is None or native_balance.amount < fee:
            raise ValueError(f"Insufficient funds to pay fee: {fee.amount} {self._gas_asset}")

    def _make_wallet(self) -> LocalWallet:
        if self.phrase or self._private_key:
            pk = PrivateKey(bytes.fromhex(self.get_private_key()))
            self._wallet = LocalWallet(pk, self._prefix)
            return self._wallet

    def get_amount_string(self, amount):
        return f"{int(amount)}{self._denom}"

    def get_denom(self, asset: Asset) -> str:
        if asset == AssetATOM:
            return COSMOS_DENOM
        else:
            return str(asset).lower()



from collections import defaultdict
from typing import NamedTuple, Optional, List


def load_logs(j_logs):
    return [TxLog.from_rpc_json(log) for log in j_logs]


class TxResponse(NamedTuple):
    height: int
    txhash: str
    data: str
    raw_log: str
    logs: list
    gas_wanted: str
    gas_used: str
    tx: Optional[dict]
    timestamp: str

    @classmethod
    def from_rpc_json(cls, j):
        return cls(
            int(j['height']),
            j['txhash'],
            j['data'],
            j['raw_log'],
            load_logs(j['logs']),
            j['gas_wanted'],
            j['gas_used'],
            j['tx'],
            j['timestamp']
        )


class TxHistoryResponse(NamedTuple):
    page_number: int
    page_total: int
    limit: int
    total: int
    tx_responses: List[TxResponse]
    txs: List[dict]

    @classmethod
    def from_rpc_json(cls, j):
        total = int(j['pagination']['total'])
        return cls(
            total=total,
            tx_responses=[TxResponse.from_rpc_json(tx) for tx in j['tx_responses']],
            txs=j['txs'],
            page_number=1,
            page_total=1,
            limit=total
        )


class TxEventAttribute(NamedTuple):
    key: str
    value: str

    @classmethod
    def from_rpc_json(cls, j):
        return cls(
            j['key'],
            j['value']
        )


class TxEvent(NamedTuple):
    type: str
    attributes: List[TxEventAttribute]

    @classmethod
    def from_rpc_json(cls, j):
        return cls(
            j['type'],
            [TxEventAttribute.from_rpc_json(attr) for attr in j['attributes']]
        )

    def find_attributes(self, key):
        return (a for a in self.attributes if a.key == key)

    def find_attr_value_first(self, key, default=None) -> Optional[str]:
        attr = next(self.find_attributes(key), None)
        return attr.value if attr else default

    @property
    def as_dict_of_list(self):
        groups = defaultdict(list)
        for attr in self.attributes:
            groups[attr.key].append(attr.value)
        return groups

    @property
    def as_dict(self):
        return {attr.key: attr.value for attr in self.attributes}


class TxLog(NamedTuple):
    msg_index: int
    log: str
    events: List[TxEvent]

    @classmethod
    def from_rpc_json(cls, j):
        return cls(
            j['msg_index'],
            j['log'],
            [TxEvent.from_rpc_json(event) for event in j['events']]
        )

    def find_events(self, event_name) -> List[TxEvent]:
        return [e for e in self.events if e.type == event_name]

    def find_event(self, event_name):
        return next((e for e in self.events if e.type == event_name), None)


class RPCResponse(NamedTuple):
    jsonrpc: str
    id: int
    result: dict

    @classmethod
    def from_rpc_json(cls, j):
        return cls(
            j['jsonrpc'],
            j['id'],
            j['result']
        )


class TxLoadException(Exception):
    pass


class TxInternalException(Exception):
    pass



import pytest

from xchainpy2_cosmos.utils import parse_cosmos_amount, parse_cosmos_amounts


def test_parse_amount():
    assert parse_cosmos_amount('0rune') == (0, 'rune')
    assert parse_cosmos_amount('1025rune') == (1025, 'rune')
    assert parse_cosmos_amount('777555444333uatom') == (777555444333, 'uatom')
    assert parse_cosmos_amount('8g') == (8, 'g')

    with pytest.raises(ValueError):
        assert parse_cosmos_amount('')

    with pytest.raises(ValueError):
        assert parse_cosmos_amount('kfehkh')

    with pytest.raises(ValueError):
        assert parse_cosmos_amount('3483')

    assert parse_cosmos_amounts('1rune,5bnb') == [(1, 'rune'), (5, 'bnb')]
    assert parse_cosmos_amounts('210ibc/0025F8A87464A471E66B234C4F93AEC5B4DA3D42D7986451A059273426290DD5,'
                                '667ibc/054892D6BB43AF8B93AAC28AA5FD7019D2C59A15DAFD6F45C1FA2BF9BDA22454,'
                                '18ibc/5CAE744C89BC70AE7B38019A1EDF83199B7E10F00F160E7F4F12BCA7A32A7EE5') == [
               (210, 'ibc/0025F8A87464A471E66B234C4F93AEC5B4DA3D42D7986451A059273426290DD5'),
               (667, 'ibc/054892D6BB43AF8B93AAC28AA5FD7019D2C59A15DAFD6F45C1FA2BF9BDA22454'),
               (18, 'ibc/5CAE744C89BC70AE7B38019A1EDF83199B7E10F00F160E7F4F12BCA7A32A7EE5'),
           ]

    # 1 amount
    assert parse_cosmos_amount('1025304930rune') == (1025304930, 'rune')



import pytest

from xchainpy2_client import TxType
from xchainpy2_cosmos import CosmosGaiaClient, COSMOS_DECIMAL
from xchainpy2_crypto import generate_mnemonic
from xchainpy2_utils import AssetATOM, Chain, NetworkType, Amount


@pytest.fixture
def cosm_client():
    return CosmosGaiaClient(phrase=generate_mnemonic())


def test_create(cosm_client):
    assert cosm_client._denom == 'uatom'
    assert cosm_client._gas_asset == AssetATOM
    assert cosm_client.chain == Chain.Cosmos
    assert cosm_client.network == NetworkType.MAINNET


def test_address():
    mn = generate_mnemonic()
    for i in range(100):
        cosm_client = CosmosGaiaClient(phrase=mn, wallet_index=i)
        address = cosm_client.get_address()
        assert cosm_client.validate_address(address)
        assert len(cosm_client.get_private_key()) == 64
        assert cosm_client.get_private_key_cosmos()
        assert len(cosm_client.get_public_key().public_key_bytes) == 33


def test_explorer(cosm_client):
    assert cosm_client.get_explorer_address_url('cosmos1mn7yrfn5ycmxsmtqvhvcac8yzyh34aknzpdkf7') \
           == 'https://www.mintscan.io/cosmos/account/cosmos1mn7yrfn5ycmxsmtqvhvcac8yzyh34aknzpdkf7'

    assert cosm_client.get_explorer_tx_url('212418055DB195796242BFDC5F26956BFBE81EF44151B73361A5BEACA989D096') \
           == 'https://www.mintscan.io/cosmos/txs/' \
              '212418055DB195796242BFDC5F26956BFBE81EF44151B73361A5BEACA989D096'


@pytest.mark.asyncio
async def test_balance(cosm_client):
    assert await cosm_client.get_balance() == []
    real_bal = await cosm_client.get_balance('cosmos1p3ucd3ptpw902fluyjzhq3ffgq4ntddac9sa3s')
    assert real_bal and isinstance(real_bal, list)
    for bal in real_bal:
        if bal.asset == AssetATOM:
            assert bal.amount.internal_amount > 100


@pytest.mark.asyncio
async def test_get_tx(cosm_client):
    tx_id = '6E5A04808C558AA35624487522B16D48275DC3D368330811ACD3020A7BDB6183'
    tx = await cosm_client.get_transaction_data(tx_id)
    assert tx
    assert tx.height == 17_945_199
    assert tx.type == TxType.TRANSFER
    assert len(tx.transfers) == 1
    tr = tx.transfers[0]
    assert tr.from_address == 'cosmos1qes70g8hc3nv7gdcyg2npjjxy4nwmxc0ege93f'
    assert tr.to_address == 'cosmos1hfdkewc9cuzdkh4f498a8e4a3hh0qpqm83s82k'
    assert tr.asset == AssetATOM
    assert tx.asset == AssetATOM
    assert tr.amount == Amount.from_base(715197, COSMOS_DECIMAL)


def test_pk_init(cosm_client):
    client2 = CosmosGaiaClient(private_key=cosm_client.get_private_key())
    assert client2.get_address() == cosm_client.get_address()
    assert client2.get_private_key() == cosm_client.get_private_key()
    assert client2.get_public_key().public_key_bytes == cosm_client.get_public_key().public_key_bytes
